/*
This file is part of the KDE libraries
    SPDX-FileCopyrightText: 2025 Ritchie Frodomar <ritchie@kde.org>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#ifndef KEYBOARDKEYSTATE_H
#define KEYBOARDKEYSTATE_H

enum class ShortcutKeyState
{
    Pressed,
    Repeated,
    Released
};

#endif